// sources_sinks.c: Demonstrate how a FIFO coordinates reads/writes
// between multiple processes.
// usage: ./a.out <delay_secs> <source_count> <sink_count>
// 
// ex: ./a.out 0.25 4 2 Runs with 4 Sources (producer) processes and 2
// Sinks (consumer) processes.  Each producer places an msg_t in a
// pipe then delays for 0.25 seconds. Each consumer extracts a msg_t
// from the pipe and delays for 0.25 seconds.
//
// This program does not terminate and can produce quite a bit of
// output so be careful
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <wait.h>
#include <assert.h>

typedef struct {
  int from;                     // write process which sent data
  int index;                    // index of message created by sender
} msg_t;

struct timespec delay = {};     // set to contain delays in main

// Write from to source fd
void write_source(int fd, int source_num){
  int count = 0;
  while(1){
    msg_t msg = {.from=source_num, .index=count};
    int nbytes = write(fd, &msg, sizeof(msg_t));
    assert(nbytes == sizeof(msg_t));
    printf("SEND %d: Proc %d sent %04d\n",
           source_num, getpid(), count);
    count++;
    nanosleep(&delay, NULL);
  }
}

// Write to the source fd
void read_source(int fd, int sink_num){
  while(1){
    msg_t msg;
    int nbytes = read(fd, &msg, sizeof(msg_t));
    assert(nbytes == sizeof(msg_t));
    printf("RECV %d: Proc %d got {from=%3d, index=%04d}\n",
           sink_num, getpid(), msg.from, msg.index);
    nanosleep(&delay, NULL);
  }
}


int main(int argc, char *argv[]){
  setvbuf(stdout, NULL, _IONBF, 0); // Turn off output buffering

  if(argc < 4){
    printf("usage: %s <delay_secs> <source_count> <sink_count>\n", argv[0]);
    exit(1);
  }

  double delay_secs = atof(argv[1]);
  delay.tv_sec  = (int) delay_secs;
  delay.tv_nsec = (delay_secs - delay.tv_sec) * 1e9;

  int n_sources = atoi(argv[2]);
  int n_sinks   = atoi(argv[3]);

  // Create FIFO to coordinate
  remove("queue.fifo");         // remove if already present
  mkfifo("queue.fifo", S_IRUSR | S_IWUSR);
  int fifo_fd = open("queue.fifo", O_RDWR);
  printf("queue.fifo fd: %d\n",fifo_fd);

  // Create writing processes
  for(int i=1; i<=n_sources; i++){
    int pid=fork();
    if(pid==0){
      write_source(fifo_fd, i);
    }
  }

  // Create reading processes
  for(int i=1; i<=n_sinks; i++){
    int pid=fork();
    if(pid==0){
      read_source(fifo_fd, i);
    }
  }

  // Not likely to ever get beyond first wait
  for(int i=0; i<n_sources+n_sinks; i++){
    wait(NULL);
  }
  return 0;
}
